import 'package:apex_socket/cipher/crypto.dart';
import 'package:flutter/cupertino.dart';
import 'package:socket_io_client/socket_io_client.dart' as IO;
import 'package:logging/logging.dart';

export 'package:socket_io_client/socket_io_client.dart';

typedef dynamic EventHandler<T>(T data);

class ApexSocket extends ChangeNotifier {
  final log = Logger('ApexSocket');

  static bool _isInitialized = false;
  static bool _isConnected = false;
  static IO.Socket socket;

  int _reconnectDelay = 2;

  /// Check if init method has been called
  bool get isInitialized => _isInitialized;

  /// Check if socket object is created and connected as well :)
  bool get isConnected => isInitialized && socket != null && _isConnected;

  /// Reconnection delay in seconds
  int get reconnectionDelay => _reconnectDelay;

  static final ApexSocket shared = ApexSocket._();

  ApexSocket._();

  static int port;
  static String url;
  static String namespace;
  static IO.OptionBuilder options;
  static bool encrypt = false;
  static String _secretKey;
  static String _publicKey;

  ApexSocket.init(String url, int port, IO.OptionBuilder options,
      {String namespace})
      : assert(url != null && Uri.parse(url).isAbsolute,
            '$url must be a valid url.'),
        assert(port == null || (port >= -1 && port <= 65535),
            '$port must be a number between -1 and 65535. or null.'),
        // assert(!_isInitialized, 'ApexSocket has been initialized once.'),
        assert(options != null, 'Options must not be null') {
    ApexSocket.url = url;
    ApexSocket.namespace = namespace;
    ApexSocket.port = port;
    ApexSocket.options = options;
    _isInitialized = true;
    notifyListeners();
  }

  ApexSocket.configure(
      {bool encrypt = false, String secretKey, String publicKey})
      : assert(
            (encrypt && secretKey != null && publicKey != null) || (!encrypt)) {
    ApexSocket.encrypt = encrypt;
    _secretKey = secretKey;
    _publicKey = publicKey;
  }

  void createSocket({
    bool recreate = false,
    EventHandler onConnect,
    EventHandler onConnectError,
    EventHandler onError,
    EventHandler onDisconnect,
    EventHandler onReconnect,
  }) {
    assert(_isInitialized, 'ApexSocket is not initialized');
    assert(options != null, 'Options must not be null');
    if (socket == null || recreate) {
      if (recreate) if (ApexSocket.socket != null) ApexSocket.socket.destroy();
      IO.Socket socket = IO.io(
          ApexSocket.port != null
              ? '${ApexSocket.url}:${ApexSocket.port}'
              : '${ApexSocket.url}' + ApexSocket.namespace != null
                  ? '/${ApexSocket.namespace}'
                  : '',
          options
              .setReconnectionDelay(_reconnectDelay)
              .setTransports(['websocket']).build());

      socket.onConnect((data) {
        _isConnected = true;
        _reconnectDelay = 2;
        notifyListeners();
        if (onConnect != null) return onConnect(data);
      });

      socket.onConnectError(onConnectError);

      socket.onError(onError);

      socket.onDisconnect((data) {
        _isConnected = false;
        notifyListeners();
        if (onDisconnect != null) return onDisconnect(data);
      });

      socket.onReconnect(onReconnect);

      socket.onReconnecting((data) {
        if (_reconnectDelay < 64) {
          _reconnectDelay *= 2;
        } else {
          _reconnectDelay = 64;
        }
        socket.io.options =
            options.setReconnectionDelay(_reconnectDelay).build();
        notifyListeners();
      });

      ApexSocket.socket = socket;
    }

    if (!ApexSocket.socket.connected)
      ApexSocket.socket.connect();
    else
      log.info('already connected');
  }

  static void emit(String event, [String data]) {
    if (encrypt) {
      final crypto = Crypto(_secretKey, _publicKey);
      crypto.encrypt(data);
    }
    ApexSocket.socket.emit(event, data);
  }

  static void emitWithAck(String event, String data,
      {Function(String data) ack}) {
    Crypto crypto;
    if (encrypt) {
      final crypto = Crypto(_secretKey, _publicKey);
      crypto.encrypt(data);
    }
    ApexSocket.socket.emitWithAck(event, data, ack: (m) {
      if (encrypt)
        ack(crypto.decrypt(m.toString()));
      else
        ack(m.toString());
    });
  }
}
